"""Exceptions for IPX800."""


class Ipx800CannotConnectError(Exception):
    """Exception to indicate an error in connection."""


class Ipx800InvalidAuthError(Exception):
    """Exception to indicate an error in authentication."""


class Ipx800RequestError(Exception):
    """Exception to indicate an error with an API request."""



"""IPX800 Virtual Output."""
from . import IPX800


class VOutput:
    """Representing an IPX800 Virtual Output."""

    def __init__(self, ipx800: IPX800, relay_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = relay_id

    @property
    async def status(self) -> bool:
        """Get status of a Virtual Output."""
        params = {"Get": "VO"}
        response = await self._ipx.request_api(params)
        return response[f"VO{self.id}"] == 1

    async def on(self) -> None:
        """Turn on a Virtual Output."""
        params = {"SetVO": self.id}
        await self._ipx.request_api(params)

    async def off(self) -> None:
        """Turn off a Virtual Output."""
        params = {"ClearVO": self.id}
        await self._ipx.request_api(params)

    async def toggle(self) -> None:
        """Toggle a Virtual Output."""
        params = {"ToggleVO": self.id}
        await self._ipx.request_api(params)



"""Get information and control a GCE IPX800v4."""
import asyncio
import socket
from time import sleep

import aiohttp
import async_timeout

from .exceptions import (
    Ipx800CannotConnectError,
    Ipx800InvalidAuthError,
    Ipx800RequestError,
)


class IPX800:
    """Class representing the IPX800 and its API."""

    def __init__(
        self,
        host: str,
        api_key: str,
        port: int = 80,
        username: str = None,
        password: str = None,
        request_retries: int = 3,
        request_timeout: int = 5,
        request_checkstatus: bool = True,
        session: aiohttp.client.ClientSession = None,
    ) -> None:
        """Init a IPX800v4 API."""
        self.host = host
        self.port = port
        self._api_key = api_key
        self._username = username
        self._password = password

        self._request_retries = request_retries
        self._request_timeout = request_timeout
        self._request_checkstatus = request_checkstatus

        self._api_url = f"http://{host}:{port}/api/xdevices.json"
        self._cgi_url = f"http://{host}:{port}/user/api.cgi"

        self._session = session
        self._close_session = False

        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._close_session = True

    async def request_api(self, params: dict) -> dict:
        """Make a request to get the IPX800 JSON API."""
        params_with_api = {"key": self._api_key}
        params_with_api.update(params)

        try:
            request_retries = self._request_retries
            while request_retries > 0:
                with async_timeout.timeout(self._request_timeout):
                    response = await self._session.get(
                        self._api_url,
                        params=params_with_api,
                    )

                if response.status:
                    content = await response.json()
                    response.close()

                    if (
                        not self._request_checkstatus
                        or content.get("status") == "Success"
                    ):
                        return content

                request_retries -= 1

            raise Ipx800RequestError("IPX800 API request error")

        except asyncio.TimeoutError as exception:
            raise Ipx800CannotConnectError(
                "Timeout occurred while connecting to IPX800."
            ) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            raise Ipx800CannotConnectError(
                "Error occurred while communicating with the IPX800."
            ) from exception

    async def request_cgi(self, params: dict) -> dict:
        """Make a request to get the IPX800 CGI API."""
        auth = None
        if self._username and self._password:
            auth = aiohttp.BasicAuth(self._username, self._password)

        try:
            request_retries = self._request_retries
            while request_retries > 0:
                with async_timeout.timeout(self._request_timeout):
                    response = await self._session.get(
                        self._cgi_url,
                        auth=auth,
                        params=params,
                    )

                if response.status == 401:
                    raise Ipx800InvalidAuthError("Auth failed on the IPX800.")

                if response.status:
                    content = await response.text()
                    response.close()
                    if not self._request_checkstatus or "Success" in content:
                        return content

                request_retries -= 1
                sleep(1)

            raise Ipx800RequestError("IPX800 API request error")

        except asyncio.TimeoutError as exception:
            raise Ipx800CannotConnectError(
                "Timeout occurred while connecting to IPX800."
            ) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            raise Ipx800CannotConnectError(
                "Error occurred while communicating with the IPX800."
            ) from exception

    async def ping(self) -> bool:
        """Return True if the IPX800 answer to API request."""
        try:
            result = await self.request_api({"Get": "R"})
            return result.get("status") == "Success"
        except Ipx800CannotConnectError:
            pass
        return False

    async def global_get(self) -> dict:
        """Get all values from the IPX800 answer."""
        values = await self.request_api({"Get": "all"})
        # add separated XPWM values if username and password set
        if self._username and self._password:
            values.update(await self.request_api({"Get": "XPWM|1-24"}))
        return values

    async def close(self) -> None:
        """Close open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self):
        """Async enter."""
        return self

    async def __aexit__(self, *_exc_info) -> None:
        """Async exit."""
        await self.close()



"""IPX800 Digital Input ."""
from . import IPX800


class DInput:
    """Representing an IPX800 Digital Input."""

    def __init__(self, ipx800: IPX800, digital_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = digital_id

    @property
    async def value(self) -> bool:
        """Get Digital Input value."""
        params = {"Get": "D"}
        response = await self._ipx.request_api(params)
        return response[f"D{self.id}"] == 1



"""IPX800 X-PWM."""
from . import IPX800

DEFAULT_TRANSITION = 500


class XPWM:
    """Representing an X-PWM channel."""

    def __init__(self, ipx800: IPX800, channel_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = channel_id

    @property
    async def status(self) -> bool:
        """Return the current X-PWM status."""
        params = {"Get": f"XPWM|{self.id}"}
        response = await self._ipx.request_api(params)
        return response[f"PWM{self.id}"] > 0

    @property
    async def level(self) -> int:
        """Return the current X-PWM level."""
        params = {"Get": f"XPWM|{self.id}"}
        response = await self._ipx.request_api(params)
        return response[f"PWM{self.id}"]

    @property
    async def level_all_channels(self) -> int:
        """Return the current X-PWM level."""
        params = {"Get": "XPWM|1-24"}
        return await self._ipx.request_api(params)

    async def on(self, time: int = DEFAULT_TRANSITION) -> None:
        """Turn on a X-PWM."""
        params = {"SetPWM": self.id, "PWMValue": "100", "PWMDelay": time}
        await self._ipx.request_cgi(params)

    async def off(self, time: int = DEFAULT_TRANSITION) -> None:
        """Turn off a X-PWM."""
        params = {"SetPWM": self.id, "PWMValue": "0", "PWMDelay": time}
        await self._ipx.request_cgi(params)

    async def toggle(self, time: int = DEFAULT_TRANSITION) -> None:
        """Toggle a X-PWM."""
        if self.status:
            self.off(time)
        else:
            self.on(time)

    async def set_level(self, level, time: int = DEFAULT_TRANSITION) -> None:
        """Turn on a X-PWM."""
        params = {"SetPWM": self.id, "PWMValue": level, "PWMDelay": time}
        await self._ipx.request_cgi(params)



"""IPX800 X-Dimmer."""
from . import IPX800

DEFAULT_TRANSITION = 500


class XDimmer:
    """Representing an X-Dimmer out."""

    def __init__(self, ipx800: IPX800, relay_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = relay_id

    @property
    async def status(self) -> bool:
        """Return the current X-Dimmer status."""
        params = {"Get": "G"}
        response = await self._ipx.request_api(params)
        return response[f"G{self.id}"]["Etat"] == "ON"

    @property
    async def level(self) -> int:
        """Return the current X-Dimmer level."""
        params = {"Get": "G"}
        response = await self._ipx.request_api(params)
        return response[f"G{self.id}"]["Valeur"]

    async def on(self, time: int = DEFAULT_TRANSITION) -> None:
        """Turn on a X-Dimmer."""
        params = {f"SetG{self.id:02}": "101", "Time": time}
        await self._ipx.request_api(params)

    async def off(self, time: int = DEFAULT_TRANSITION) -> None:
        """Turn off a X-Dimmer."""
        params = {f"SetG{self.id:02}": "0", "Time": time}
        await self._ipx.request_api(params)

    async def toggle(self, time: int = DEFAULT_TRANSITION) -> None:
        """Toggle a X-Dimmer."""
        if await self.status:
            await self.off(time)
        else:
            await self.on(time)

    async def set_level(self, level: int, time: int = DEFAULT_TRANSITION) -> None:
        """Turn on a X-Dimmer on a specific level."""
        params = {f"SetG{self.id:02}": level, "Time": time}
        await self._ipx.request_api(params)



"""IPX800 X-THL."""
from . import IPX800


class XTHL:
    """Representing an IPX800 X-THL."""

    def __init__(self, ipx800: IPX800, xthl_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = xthl_id

    @property
    async def temp(self) -> float:
        """Get temperature of the X-THL."""
        params = {"Get": "XTHL"}
        response = await self._ipx.request_api(params)
        return response[f"THL{self.id}-TEMP"]

    @property
    async def hum(self) -> float:
        """Get humidity level of the X-THL."""
        params = {"Get": "XTHL"}
        response = await self._ipx.request_api(params)
        return response[f"THL{self.id}-HUM"]

    @property
    async def lum(self) -> int:
        """Get luminosity level of the X-THL."""
        params = {"Get": "XTHL"}
        response = await self._ipx.request_api(params)
        return response[f"THL{self.id}-LUM"]



"""IPX800 X-4VR."""
from . import IPX800


class X4VR:
    """Representing an X-4VR output."""

    def __init__(self, ipx800: IPX800, ext_id: int, vr_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.ext_id = ext_id
        self.vr_id = vr_id
        self.vr_number = (ext_id - 1) * 4 + vr_id

    @property
    async def status(self) -> bool:
        """Return the current cover status."""
        params = {"Get": f"VR{self.ext_id}"}
        response = await self._ipx.request_api(params)
        return response[f"VR{self.ext_id}-{self.vr_id}"] < 100

    @property
    async def level(self) -> int:
        """Return the current cover level."""
        params = {"Get": f"VR{self.ext_id}"}
        response = await self._ipx.request_api(params)
        return 100 - int(response[f"VR{self.ext_id}-{self.vr_id}"])

    async def on(self) -> None:
        """Open cover."""
        params = {f"SetVR{self.vr_number:02}": "0"}
        await self._ipx.request_api(params)

    async def off(self) -> None:
        """Close cover."""
        params = {f"SetVR{self.vr_number:02}": "100"}
        await self._ipx.request_api(params)

    async def stop(self) -> None:
        """Stop cover."""
        params = {f"SetVR{self.vr_number:02}": "101"}
        await self._ipx.request_api(params)

    async def set_level(self, level: int) -> None:
        """Set cover level."""
        params = {f"SetVR{self.vr_number:02}": str(100 - level)}
        await self._ipx.request_api(params)



"""IPX800 X-4FP."""
from . import IPX800


class X4FP:
    """Representing an X-4FP output."""

    def __init__(self, ipx800: IPX800, ext_id: int, zone_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.ext_id = ext_id
        self.zone_id = zone_id
        self.fp_number = (ext_id - 1) * 4 + zone_id

    @property
    async def status(self) -> bool:
        """Return the current FP status."""
        params = {"Get": "FP"}
        response = await self._ipx.request_api(params)
        return response[f"FP{self.ext_id} Zone {self.zone_id}"]

    async def set_mode(self, mode) -> None:
        """Set FP mode."""
        params = {f"SetFP{self.fp_number:02}": mode}
        await self._ipx.request_api(params)

    async def set_mode_all(self, mode) -> None:
        """Set FP mode for all zones."""
        params = {"SetFP00": mode}
        await self._ipx.request_api(params)



"""IPX800 relay."""
from . import IPX800


class Relay:
    """Representing an IPX800 relay."""

    def __init__(self, ipx800: IPX800, relay_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = relay_id

    @property
    async def status(self) -> bool:
        """Return the current relay status."""
        params = {"Get": "R"}
        response = await self._ipx.request_api(params)
        return response[f"R{self.id}"] == 1

    async def on(self) -> None:
        """Turn on a relay."""
        params = {"SetR": self.id}
        await self._ipx.request_api(params)

    async def off(self) -> None:
        """Turn off a relay."""
        params = {"ClearR": self.id}
        await self._ipx.request_api(params)

    async def toggle(self) -> None:
        """Toggle a relay."""
        params = {"ToggleR": self.id}
        await self._ipx.request_api(params)



"""Asynchronous Python client for the IPX800 v4 API."""

from .ipx800 import (
    IPX800,
    Ipx800CannotConnectError,
    Ipx800InvalidAuthError,
    Ipx800RequestError,
)

from .ainput import AInput
from .vainput import VAInput
from .dinput import DInput
from .relay import Relay
from .vinput import VInput
from .voutput import VOutput
from .x4fp import X4FP
from .x4vr import X4VR
from .xdimmer import XDimmer
from .xpwm import XPWM
from .xthl import XTHL

__all__ = [
    "IPX800",
    "Ipx800CannotConnectError",
    "Ipx800InvalidAuthError",
    "Ipx800RequestError",
    "Relay",
    "XDimmer",
    "XPWM",
    "AInput",
    "VAInput",
    "DInput",
    "VInput",
    "VOutput",
    "X4FP",
    "X4VR",
    "XTHL",
]



"""IPX800 Analog Input."""
from . import IPX800


class AInput:
    """Representing an IPX800 Analog Input."""

    def __init__(self, ipx800: IPX800, analog_id: int) -> None:
        """Initialize object."""
        self._ipx = ipx800
        self.id = analog_id

    @property
    async def value(self) -> float:
        """Get Analog Input value."""
        params = {"Get": "A"}
        response = await self._ipx.request_api(params)
        return response[f"A{self.id}"]
